import os
import yadageschemas
import json
import logging

from . import workflow_loader as workflow_loader

from .handlers.utils import handler_decorator
from .state_providers import state_provider_from_string
from .wflowstate import load_model_fromstring
from .controllers import setup_controller
from .wflow import YadageWorkflow

log = logging.getLogger(__name__)

handlers, creator = handler_decorator()


log = logging.getLogger(__name__)

@creator('local')
def local_workflows(
    metadir,
    workflow = None,
    initdata = None,
    toplevel = os.getcwd(),
    dataarg = None,
    dataopts = None,
    wflowopts = None,
    workflow_json = None,
    modelsetup = 'inmem',
    modelopts = None,
    controller = 'frommodel',
    ctrlopts = None,
    validate = True,
    schemadir = yadageschemas.schemadir):
    '''
    load workflow from spec and initialize it

    :param workflow: the workflow spec source
    :param toplevel: base URI against which to resolve JSON references in the spec
    :param initdata: initialization data for workflow

    prepares initial workflow object and returns controller
    '''
    rootprovider = state_provider_from_string(dataarg, dataopts)

    if not workflow_json and not workflow:
        raise RuntimeError('need to provide either direct workflow spec or source to load from')

    if workflow_json:
        if validate: workflow_loader.validate(workflow_json)
    else:
        workflow_json = workflow_loader.workflow(
            workflow,
            toplevel=toplevel,
            schemadir=schemadir,
            validate=validate
        )

    with open('{}/yadage_template.json'.format(metadir), 'w') as f:
        json.dump(workflow_json, f)
    workflowobj = YadageWorkflow.createFromJSON(workflow_json, rootprovider)
    if initdata:
        wflowopts = wflowopts or {}
        discover = wflowopts.setdefault('discover',True)
        relative = wflowopts.setdefault('relative',True)
        log.info('initializing workflow with initdata: %s discover: %s relative: %s',initdata, discover, relative)
        workflowobj.view().init(
            initdata, rootprovider, discover = discover, relative = relative
        )
    else:
        log.info('no initialization data')

    model = load_model_fromstring(modelsetup,modelopts,workflowobj)
    return setup_controller(
        model = model,
        controller = controller, ctrlopts = ctrlopts,
    )



import copy
import hashlib
import json
import logging
import os
import uuid

import jq
import jsonpointer
import yaml

from yadageschemas.utils import WithJsonRefEncoder

class outputReference(object):

    def __init__(self, stepid, pointer):
        self.stepid = stepid
        self.pointer = pointer

    def __repr__(self):
        return 'outputReference {}#{}'.format(self.stepid, self.pointer.path)

    #(de-)serialization
    @classmethod
    def fromJSON(cls, data):
        return cls(data['stepid'], jsonpointer.JsonPointer(data['pointer_path']))

    def json(self):
        return {
            'stepid': self.stepid,
            'pointer_path': self.pointer.path
        }


log = logging.getLogger(__name__)

DEFAULT_ID_METHOD = 'uuid'


def json_hash(jsonable):
    '''
    returns a content-based hash of a json-able structure (serialized with WithJsonRefEncoder)

    :param jsonable: a json-serializable object
    :return: the hash
    '''
    the_hash  = hashlib.sha1(json.dumps(jsonable, cls=WithJsonRefEncoder, sort_keys=True).encode('utf-8')).hexdigest()
    return the_hash


def get_id_fromjson(jsonobject, method = DEFAULT_ID_METHOD):
    method = os.environ.get('YADAGE_ID_METHOD', method)
    if method == 'uuid':
        return str(uuid.uuid4())
    elif method == 'jsonhash':
        return json_hash(jsonobject)
    else:
        raise NotImplementedError(
            'unkown id generation method {}'.format(method))

def get_obj_id(obj_with_json_method, method = DEFAULT_ID_METHOD):
    return get_id_fromjson(obj_with_json_method.json(), method)

def process_refs(x, dag):
    if type(x) == dict:
        for k, v in x.items():
            x[k] = process_refs(v, dag)
        return x
    elif type(x) == list:
        for i, e in enumerate(x):
            x[i] = process_refs(e, dag)
        return x
    elif type(x) == outputReference:
        return x.pointer.resolve(dag.getNode(x.stepid).result)
    else:
        return x

def leaf_iterator_jsonlike(jsonlike, path = None):
    ''''
    :param jsonlike: a jsonlike object (i.e. nested lists/arrays: leaf-types must not be JSONable)
    :return: iterator that yields (JSON path, leaf value) tuples
    '''
    path = path or []
    if type(jsonlike) == list:
        for i,x in enumerate(jsonlike):
            thispath = copy.deepcopy(path)
            thispath.append(i)
            for leaf in leaf_iterator_jsonlike(x, path = thispath):
                yield leaf
    elif type(jsonlike) == dict:
        for k,v in jsonlike.items():
            thispath = copy.deepcopy(path)
            thispath.append(k)
            for leaf in leaf_iterator_jsonlike(v, path = thispath):
                yield leaf
    else:
        yield jsonpointer.JsonPointer.from_parts(path),jsonlike


def process_jsonlike(jsonlike, jq_obj_selector, callback):
    wflowrefs = [jsonpointer.JsonPointer.from_parts(x[1:]) for x in jq.jq(
            'paths(if objects then {} else false end)'.format(jq_obj_selector)
            ).transform({'value': jsonlike}, multiple_output = True
    )]
    for wflowref in wflowrefs:
        value = callback(wflowref.resolve(jsonlike))
        if wflowref.path == '':
            return value
        else:
            wflowref.set(jsonlike,value)
    return jsonlike


def options_from_eqdelimstring(opts):
    options = {}
    for x in opts:
        key, value = x.split('=',1)
        options[key] = yaml.load(value)
    return options

def getinit_data(initfiles, parameters):
    '''
    get initial data from both a list of files and a list of 'pname=pvalue'
    strings as they are passed in the command line <pvalue> is assumed to be a
    YAML parsable string.
    '''

    initdata = {}
    for initfile in initfiles:
        log.debug('loading initialization data from file %s',initfile)
        initdata.update(**yaml.load(open(initfile)))

    initdata.update(**options_from_eqdelimstring(parameters))
    return initdata

def setupbackend_fromstring(backend, backendopts = None):
    backendopts = backendopts or {}
    log.debug('setting up backend %s with opts %s', backend, backendopts)
    import yadage.backends.packtivitybackend as pb
    return pb.PacktivityBackend(
            packtivity_backendstring = backend,
            backendopts = backendopts
    )

def init_stage_spec(parameters, discover, used_inputs, name, nodename = None, relative = False):
    return {
        'name': name,
        'dependencies': {
            "dependency_type": "jsonpath_ready",
            "expressions": []
        },
        'scheduler': {
             'scheduler_type': 'init-stage',
             'parameters': parameters,
             'inputs':   used_inputs,
             'nodename': nodename,
             'step': {
                'process': None,
                'environment': None,
                'publisher': {
                    'publisher_type': 'fromparjq-pub',
                    'script': '.',
                    'tryExact': True,
                    'glob': discover,
                    'relative_paths': relative
                }
            }
        }
    }

def stages_in_scope(workflow,scope):
    return jsonpointer.JsonPointer(scope).resolve(workflow.bookkeeping)['_meta']['stages']

def rule_steps_indices(workflow):
    rule_to_steps_index = {}
    steps_to_rule_index = {}
    rule_to_subscopes_index = {}
    for rule in workflow.rules + workflow.applied_rules:
        path = '/'.join([rule.offset, rule.rule.name])
        p = jsonpointer.JsonPointer(path)
        try:
            steps_of_rule = [x['_nodeid'] for x in p.resolve(workflow.stepsbystage) if '_nodeid' in x]
        except jsonpointer.JsonPointerException:
            steps_of_rule = []

        try:
            a = p.resolve(workflow.stepsbystage)
            subscopes_of_rule = [
                # ['{}/{}'.format(x['_offset'],substage) for substage in x.keys() if not substage== '_offset' ]
                x['_offset']
                for x in a if '_offset' in x
            ]
        except jsonpointer.JsonPointerException:
            subscopes_of_rule = []

        rule_to_steps_index[rule.identifier] = steps_of_rule
        rule_to_subscopes_index[rule.identifier] = subscopes_of_rule
        for step in steps_of_rule:
            steps_to_rule_index[step] = rule.identifier
    return rule_to_steps_index, steps_to_rule_index, rule_to_subscopes_index

def advance_coroutine(coroutine):
    try:
        return coroutine.next()
    except AttributeError:
        return coroutine.__next__()

def prepare_meta(metadir, accept=False):
    '''
    prepare workflow meta-data directory

    :param metadir: the meta-data directory name
    :param accept: whether to accept an existing metadata directory
    '''
    if os.path.exists(metadir):
        if not accept:
            raise RuntimeError("yadage meta directory %s exists. Allow overwrite by using the command line option  --accept-metadir" % metadir)
    else:
        os.makedirs(metadir)

def pointerize(data, asref=False, stepid=None):
    def callback(p):
        return outputReference(stepid, p) if asref else {'$wflowpointer': {'step': stepid,'result': p.path}} if stepid else p.path
    return data.asrefs(callback = callback)



import logging

import networkx as nx

import yadage.utils as utils
import adage.nodestate
log = logging.getLogger(__name__)

def reset_node_state(node):
    node.submit_time = None
    node.ready_by_time = None
    node.resultproxy = None
    node.backend = None
    node.expected_result = None
    node._result = None
    node._state = adage.nodestate.DEFINED

def reset_step(workflow, step):
    s = workflow.dag.getNode(step)
    log.debug('resetting DAG node %s/%s:%s',s.task.metadata['wflow_offset'],s.task.metadata['wflow_stage'],s.task.metadata['wflow_stage_node_idx'])
    reset_node_state(s)
    try:
        s.task.state.reset()
    except AttributeError:
        pass

def remove_rule(workflow, ruleid):
    assert ruleid in [x.identifier for x in workflow.rules]
    r = workflow.view().getRule(identifier = ruleid)
    log.debug('>>> removing rule %s/%s',r.offset,r.rule.name)
    workflow.rules.remove(r)
    view = workflow.view(r.offset)
    stagemeta = view.bookkeeper['_meta']['stages']
    view.bookkeeper['_meta']['stages'] = [x for x in stagemeta if not x==ruleid]

def remove_rules(workflow, ruleids):
    for r in ruleids:
        remove_rule(workflow, r)

def undo_rule(workflow, ruleid):
    r2s, s2r, r2subscopes  = utils.rule_steps_indices(workflow)

    if not ruleid in [r.identifier for r in workflow.applied_rules]:
        ruleobj = workflow.view().getRule(identifier = ruleid)
        log.debug('rule %s/%s not in list of applied rules. possibly already undone during recursion.',ruleobj.offset,ruleobj.rule.name)
        return

    downstream_nodes_rules = list(set([s2r[s] for s in collective_downstream(workflow, r2s[ruleid])]))


    r = workflow.view().getRule(identifier = ruleid)
    log.debug('undoing %s/%s. downstream_nodes_rules: %s', r.offset, r.rule.name, len(downstream_nodes_rules))

    assert ruleid not in downstream_nodes_rules
    #TODO: not an issue now, but if stages schedule linked nodes we might
    #we might see downstream steps that are part of the same rule

    if not downstream_nodes_rules:
        idx   = workflow.applied_rules.index(r)
        stepids = r2s[ruleid]
        steps = [workflow.dag.getNode(nid) for nid in stepids]

        # reset states of the step
        reset_steps(workflow,stepids)
        #
        # delete the steps
        for n in steps:
            log.debug('remove DAG node %s/%s:%s',n.task.metadata['wflow_offset'],n.task.metadata['wflow_stage'],n.task.metadata['wflow_stage_node_idx'])
            workflow.dag.removeNode(n)

        for subscope in r2subscopes[ruleid]:
            subrules = utils.stages_in_scope(workflow,subscope)
            log.debug('remove subscope %s with %s subrules', subscope, len(subrules))
            for subruleidx,subrule in enumerate(subrules):
                subruleobj = workflow.view().getRule(identifier = subrule)
                log.debug('undo sub DAG rule %s/%s', subruleobj.offset, subruleobj.rule.name)
                undo_rule(workflow,subrule)

            log.debug('removing %s subrules', len(subrules))
            for subruleidx,subrule in enumerate(subrules):
                subruleobj = workflow.view().getRule(identifier = subrule)
                log.debug('remove sub DAG rule %s/%s', subruleobj.offset, subruleobj.rule.name)
                remove_rule(workflow,subrule)
        #remove rule from applied list
        r = workflow.applied_rules.pop(idx)

        #reset all index data for stage
        workflow.view(r.offset).steps.pop(r.rule.name)

        #remove steps and stage from bookkeeping
        for s in stepids:
            workflow.view(r.offset).bookkeeper['_meta']['steps'].remove(s)

        workflow.view(r.offset).bookkeeper['_meta']['stages'].remove(ruleid)

        #re-append the rule
        log.debug('re-appened {}/{}'.format(r.offset,r.rule.name))

        # workflow.steps[stage][-1]['_offset'] = offset
        newid = workflow.view(r.offset).addRule(r.rule, identifier = r.identifier)
        assert newid == r.identifier

        log.debug('undo any rules that would not be applicable now')
        for rule in workflow.applied_rules:
            if not rule.applicable(workflow):
                ruleobj = workflow.view().getRule(identifier = rule.identifier)
                log.debug('rule would not be appilcable in current state so undo >> %s/%s', ruleobj.offset, ruleobj.rule.name)
                undo_rule(workflow,rule.identifier)

        return

    log.debug('to undo a rule, we need to undo rules responsible for any of the downstream nodes of the rules of this stage')
    while True:
        if not downstream_nodes_rules:
            break
        for r in downstream_nodes_rules:
            log.debug('undoing a downstream rule')
            undo_rule(workflow,r)
            log.debug('undone a downstream rule')
        r2s, s2r, _   = utils.rule_steps_indices(workflow)
        downstream_nodes_rules = list(set([s2r[s] for s in collective_downstream(workflow, r2s[ruleid])]))
        log.debug('re-asses if there are still any downstream rules: {}'.format(len(downstream_nodes_rules)))
    log.debug('undo rule (now without any downstream rules)')
    undo_rule(workflow,ruleid)



def undo_rules(workflow, ruleids):
    for r in ruleids:
        undo_rule(workflow, r)

def reset_steps(workflow, steps):
    for s in steps:
        reset_step(workflow, s)

def collective_downstream(workflow, steps):
    downstream = set()
    for step in steps:
        for x in nx.descendants(workflow.dag, step):
            downstream.add(x)
    return list(downstream)



import os
# import datetime
# import time

import jsonpointer

import adage.node
import adage.serialize

from packtivity import datamodel

from packtivity.backendutils import load_proxy
from .tasks import outputReference, packtivity_task

class YadageNode(adage.node.Node):
    '''
    Node object for yadage that extends the default with
    the ability to have prepublished results
    '''

    def __init__(self, name, task, identifier=None, result = None):
        super(YadageNode, self).__init__(name, task, identifier, result = result)
        self.expected_result = None

    def __repr__(self):
        # lifetime = datetime.timedelta(seconds = (time.time() - self.define_time))
        # runtime = None
        if self.state != adage.nodestate.DEFINED:
            pass
            # referencetime = time.time() if not self.ready() else self.ready_by_time
            # runtime = datetime.timedelta(seconds = (referencetime - self.submit_time))
        return '<{}/{}:{}|{}|{}>'.format(
            self.task.metadata.get('wflow_offset','-'),
            self.task.metadata.get('wflow_stage','-'),
            self.task.metadata.get('wflow_stage_node_idx','-'),
            str(self.state).lower(),
            'known' if self.has_result() else 'unknown'
        )

    def has_result(self):
        return (self.expected_result is not None) or self.successful()

    @property
    def result(self):
        if self.expected_result is not None and 'YADAGE_IGNORE_PREPUBLISHING' not in os.environ:
            if self.ready() and self.successful():
                sanity =  super(YadageNode, self).result.json() == self.expected_result.json()
                if not sanity:
                    raise RuntimeError('prepublished and actual result differ:\n result:\n{}\n prepub:{}'.format(
                        super(YadageNode, self).result.json(),self.expected_result.json())
                )
            return self.expected_result
        return super(YadageNode, self).result

    def readfromresult(self,pointerpath, trackinputs = None, failsilently = False):
        if not self.has_result():
            if failsilently: return None
            raise RuntimeError('attempt')
        pointer = jsonpointer.JsonPointer(pointerpath)
        if trackinputs is not None:
            trackinputs.append(outputReference(self.identifier,pointer))
        v = self.result.resolve_ref(pointer)
        return v

    def json(self):
        json_or_nil = lambda x: None if x is None else x.json()
        d = adage.serialize.node_to_json(self,json_or_nil,json_or_nil)
        d['result'] = json_or_nil(self.result if self.has_result() else None)
        return d

    @classmethod
    def fromJSON(cls, data, deserialization_opts = None):
        if data['task']['type'] == 'packtivity_task':
            task   = packtivity_task.fromJSON(data['task'], deserialization_opts)
            result = datamodel.create(data['result'],getattr(task.state,'datamodel',{})) if data['result'] else None
            instance = cls(data['name'], task, data['id'],result)

            adage.serialize.set_generic_data(instance,data)
            instance.resultproxy = load_proxy(data['proxy'],deserialization_opts, best_effort_backend = False) if data['proxy'] else None
            return instance
        else:
            raise RuntimeError('unknown task type',data['task']['type'])



import json
import logging
import shutil
import tempfile
import click
import yaml
import os
import base64
import glob

from .handlers.expression_handlers import handlers as exh
from .utils import process_refs
from .wflow import YadageWorkflow


def printRef(ref, dag, indent=''):
    click.secho('{}name: {} position: {}, value: {}, id: {}'.format(
        indent,
        dag.getNode(ref.stepid).name,
        ref.pointer.path,
        ref.pointer.resolve(dag.getNode(ref.stepid).result),
        ref.stepid
    ),
        fg='cyan')

def wflow_with_trivial_backend(instance):

    stateopts = {}
    wflow = YadageWorkflow.fromJSON(json.load(open(instance)),stateopts)
    return wflow

@click.group()
def utilcli():
    pass


@utilcli.command()
@click.argument('instance')
@click.argument('selection')
@click.option('--viewscope', default = '')
@click.option('-v', '--verbosity', default='INFO')
def testsel(instance, selection,verbosity,viewscope):


    logging.basicConfig(level=getattr(logging, verbosity), format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')


    wflow = wflow_with_trivial_backend(instance)

    selresult = exh['stage-output-selector'](wflow.view(viewscope), yaml.load(selection))

    if not selresult:
        click.secho('Bad selection {}'.format(selresult), fg='red')
        return

    click.secho(json.dumps(
        process_refs(selresult, wflow.dag),
        sort_keys=True,
        indent=4,
        separators=(',', ': ')),
        fg='green'
    )


@utilcli.command()
@click.argument('instance')
@click.argument('vizpdf')
@click.option('--viewscope', default = '')
@click.option('-v', '--verbosity', default='INFO')
def viz(instance, vizpdf,viewscope,verbosity):
    logging.basicConfig(level=getattr(logging, verbosity), format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    import yadage.visualize as visualize
    wflow = wflow_with_trivial_backend(instance)

    dirpath = tempfile.mkdtemp()
    visualize.write_prov_graph(dirpath, wflow, scope = viewscope)
    shutil.copy('{}/yadage_workflow_instance.pdf'.format(dirpath), vizpdf)
    shutil.rmtree(dirpath)

@utilcli.group()
def k8s():
    pass

@k8s.command()
@click.option('--hostname', default = 'docker-for-desktop')
@click.option('--path', default = None)
def create_state(hostname,path):
    pvc_name = 'yadagedata'
    sc_name = 'local-storage'
    path_base = path or os.getcwd()
    size = '1G'
    kubeyaml = '''\
kind: PersistentVolumeClaim
apiVersion: v1
metadata:
  name: {pvc_name}
spec:
  accessModes:
    - ReadWriteMany
  resources:
    requests:
      storage: {size}
  storageClassName: {sc_name}
---
kind: StorageClass
apiVersion: storage.k8s.io/v1
metadata:
  name: {sc_name}
provisioner: kubernetes.io/no-provisioner
volumeBindingMode: WaitForFirstConsumer
---
apiVersion: v1
kind: PersistentVolume
metadata:
  name: yadage-pv
spec:
  capacity:
    storage: {size}
  volumeMode: Filesystem
  accessModes:
  - ReadWriteMany
  persistentVolumeReclaimPolicy: Delete
  storageClassName: {sc_name}
  local:
    path: {path_base}
  nodeAffinity:
    required:
      nodeSelectorTerms:
      - matchExpressions:
        - key: kubernetes.io/hostname
          operator: In
          values:
          - {hostname}
'''.format(
        pvc_name = pvc_name,
        sc_name = sc_name,
        base_path = path_base,
        size = size,
        path_base = path_base,
        hostname = hostname
    )
    click.echo(kubeyaml)


@k8s.command()
@click.option('--path')
def create_secrets(path):
    files = glob.glob('{}/*'.format(path or os.environ['PACKTIVITY_AUTH_LOCATION']))
    data = {
      os.path.basename(f):base64.b64encode(open(f,'rb').read()).decode('ascii') for f in files
    }
    secrets = '''\
apiVersion: v1
data: {}
kind: Secret
metadata:
  name: hepauth
  namespace: default
type: Opaque
'''.format(json.dumps(data))
    click.echo(secrets)

if __name__ == '__main__':
    utilcli()



import json
import logging

from .wflow import YadageWorkflow
from .handlers.utils import handler_decorator

log = logging.getLogger(__name__)

modelhandlers, statemodel = handler_decorator()

@statemodel('inmem')
def inmem_model(modelsetup, modelopts, initmodel):
    return initmodel

@statemodel('filebacked')
def filebacked_model(modelsetup, modelopts, initmodel):
    filename = modelsetup.split(':')[-1]
    model   = FileBackedModel(
        filename = filename,
        initmodel = initmodel,
        deserialization_opts = modelopts
    )
    return model

@statemodel('mongo')
def mongo_model(modelsetup, modelopts, initmodel):
    model = MongoBackedModel(
        initmodel = initmodel, deserialization_opts = modelopts
    )
    return model

def load_model_fromstring(modelsetup,modelopts = None,initmodel = None):
    modelopts = modelopts or {}
    for k in modelhandlers.keys():
        if modelsetup.startswith(k):
            return modelhandlers[k](modelsetup,modelopts,initmodel)
    raise RuntimeError('unknown state model %s' % modelsetup)

class MongoBackedModel(object):
    '''
    model that holds the workflow state in a MongoDB database.
    '''
    def __init__(self, deserialization_opts = None, connect_string = 'mongodb://localhost:27017/', initmodel = None, wflowid = None):
        from pymongo import MongoClient
        from bson.objectid import ObjectId
        self.deserialization_opts = deserialization_opts
        self.client = MongoClient(connect_string)
        self.db = self.client.wflowdb
        self.collection = self.db.workflows
        if initmodel:
            insertion = self.collection.insert_one(initmodel.json())
            self.wflowid = insertion.inserted_id
            log.info('created new workflow object with id %s', str(self.wflowid))
        if wflowid:
            self.wflowid = ObjectId(wflowid)

    def commit(self, data):
        '''
        :param data: data to commit to disk. needs to have '.json()' method
        return: None
        '''
        self.collection.replace_one({'_id':self.wflowid}, data.json())

    def load(self):
        '''
        param load: load data from database
        return: yadage workflow object
        '''
        jsondata = self.collection.find_one({'_id':self.wflowid})
        return YadageWorkflow.fromJSON(jsondata,self.deserialization_opts)

class FileBackedModel(object):
    '''
    model that holds data on disk in a JSON file
    '''
    def __init__(self, filename, deserialization_opts = None, initmodel = None):
        self.filename = filename
        self.deserialization_opts = deserialization_opts
        if initmodel:
            self.commit(initmodel)

    def commit(self, data):
        '''
        :param data: data to commit to disk. needs to have '.json()' method
        '''
        log.debug('committing model')
        jsondata = data.json()

        with open(self.filename,'w') as statefile:
            json.dump(jsondata, statefile)

    def load(self):
        '''
        :return: the adage workflow object holding rules and the graph
        '''
        log.debug('loading model')
        with open(self.filename) as statefile:
            jsondata = json.load(statefile)
            return YadageWorkflow.fromJSON(jsondata,self.deserialization_opts)



import json


def snapshot(workflow, wflowfile):
    '''
    snapshots the state of the workflow into two files:
        - a file holding the pure state of the workflow
        - a file holding the backend state, i.e. results to specific tasks
    '''
    with open(wflowfile, 'w') as wfile:
        json.dump(workflow.json(), wfile)



from packtivity.backendutils import proxyloader, load_proxy
from yadage.backends.trivialbackend import TrivialProxy

class CachedProxy(object):
    def __init__(self,proxy,cacheid):
        self.proxy = proxy
        self.cacheid = cacheid

    def details(self):
        return {}

    def json(self):
        return {'proxyname': 'CachedProxy', 'proxy': self.proxy.json(), 'cacheid': self.cacheid}

    @classmethod
    def fromJSON(cls,data,deserialization_opts = None):
        deserialization_opts = deserialization_opts or {}
        return cls(load_proxy(data['proxy'],deserialization_opts), data['cacheid'],best_effort_backend = False)

@proxyloader('CachedProxy')
def cache_loader(jsondata, deserialization_opts = None, best_effort_backend = False):
    proxy = CachedProxy.fromJSON(jsondata,deserialization_opts)
    if best_effort_backend:
        raise NotImplementedError('nope')
    return proxy

@proxyloader('TrivialProxy')
def trivial_loader(jsondata, deserialization_opts = None, best_effort_backend = False):
    proxy = TrivialProxy.fromJSON(jsondata)
    if best_effort_backend:
        raise NotImplementedError('nope')
    return proxy



import logging
import os
import shutil
from six import string_types
import zipfile
from packtivity.statecontexts import load_state
from packtivity.statecontexts.posixfs_context import LocalFSState
import fnmatch

log = logging.getLogger(__name__)


import requests

try:	
    # For Python 3.0 and later	
    from urllib.request import urlopen	
except ImportError:	
    # Fall back to Python 2's urllib2	
    from urllib2 import urlopen
    
def download_file(url,local_filename):
    if not 'http' in url:
        f = urlopen(url)
        with open(local_filename,'wb') as lf:
            lf.write(f.read())
    else:
        h = {} if not 'YADAGE_INIT_TOKEN' in os.environ else {'PRIVATE-TOKEN': os.environ['YADAGE_INIT_TOKEN']}
        r = requests.get(url, stream=True, headers = h)
        with open(local_filename, 'wb') as f:
            shutil.copyfileobj(r.raw, f)

def prepare_workdir_from_archive(initdir, inputarchive, match = None):
    url = inputarchive
    os.makedirs(initdir)
    localzipfile = '{}/.yadage_inputarchive.zip'.format(initdir)
    download_file(url,localzipfile)

    if not match:
        with zipfile.ZipFile(localzipfile) as zf:
            zf.extractall(path=initdir)
    else:
        assert match.endswith('/')
        with zipfile.ZipFile(localzipfile) as zf:
            filtered = fnmatch.filter(zf.namelist(),match)                
            assert len(filtered)==1
            basepath = filtered[0]
            for x in zf.namelist():
                if x!= basepath and basepath in x:
                    target = x.replace(basepath,'')
                    if target.endswith('/'):
                        os.makedirs(os.path.join(initdir,target))
                    else:
                        with open(os.path.join(initdir,target),'wb') as f:
                            f.write(zf.read(x))
    os.remove(localzipfile)

def _merge_states(lhs,rhs):
    return LocalFSState(lhs.readwrite + rhs.readwrite,lhs.readonly + rhs.readonly)

class LocalFSProvider(object):
    def __init__(self, *base_states, **kwargs):
        base_states = list(base_states)
        self.nest = kwargs.get('nest', True)
        self.ensure = kwargs.get('ensure')
        self.init_states = kwargs.get('init_states')
        self.sub_inits = kwargs.get('sub_init_states',{})

        first = base_states.pop()
        assert first

        self.base = first

        while base_states:
            next_state = base_states.pop()
            if not next_state:
                continue
            self.base = _merge_states(self.base,next_state)

    def new_provider(self,name, init_states = None):
        new_base_ro = self.base.readwrite + self.base.readonly
        new_base_rw = [os.path.join(self.base.readwrite[0],name)]

        sub_init    = self.sub_inits.get(name,{}).get('_data')
        sub_inits   = [sub_init] if sub_init else []
        init_states = init_states or []

        newsinits = {k:v for k,v in self.sub_inits.get(name,{}).items() if k!='_data'}
        return LocalFSProvider(LocalFSState(new_base_rw,new_base_ro),
                               nest = self.nest,
                               ensure = self.ensure,
                               init_states = init_states + sub_inits,
                               sub_init_states = newsinits)


    def new_state(self,name, dependencies, readonly = False):
        '''
        creates a new context from an existing context.

        if subdir is True it declares a new read-write nested under the old
        context's read-write and adds all read-write and read-only locations
        of the old context as read-only. This is recommended as it makes rolling
        back changes to the global state made in this context easy.

        else the same readwrite/readonly configuration as the parent context is used
        '''

        readwrites = []
        if not readonly:
            readwrites = ['{}/{}'.format(self.base.readwrite[0],name)] if self.nest else self.base.readwrite

        log.debug('new state is: rw: %s', readwrites)
        state_identifier = name.replace('/','_') # replace in case name is nested path
        newstate = LocalFSState(readwrite = readwrites, identifier = state_identifier, dependencies = dependencies)

        if not readonly and self.ensure:
            newstate.ensure()

        return newstate

    def json(self):
        def subinit_json(subinits):
            def handle(k,v):
                if k=='_data':
                    return v.json() if v else None
                return {kk:handle(kk,vv) for kk,vv in v.items()}
            return {k: handle(k,v) for k,v in subinits.items()}
        d = {
            'state_provider_type': 'localfs_provider',
            'base_state': self.base.json(),
            'init_states': [s.json() for s in self.init_states] if self.init_states else [],
            'nest': self.nest,
            'ensure': self.ensure,
            'sub_init_states': subinit_json(self.sub_inits)
        }
        return d

    @classmethod
    def fromJSON(cls,jsondata, deserialization_opts):
        def subinit_load(subinit_data):
            def handle(k,v):
                if k == '_data':
                    return load_state(v,deserialization_opts) if v else None
                return {kk:handle(kk,vv) for kk,vv in v.items()}
            return {k:handle(k,v) for k,v in subinit_data.items()}

        instance = cls(
            load_state(jsondata['base_state'],deserialization_opts),
            nest = jsondata['nest'],
            ensure = jsondata['ensure'],
            init_states = [load_state(x,deserialization_opts) for x in jsondata['init_states']],
            sub_init_states = subinit_load(jsondata['sub_init_states'])
        )
        return instance

def setup_provider(dataarg, dataopts):
    workdir   = dataarg.split(':',2)[1]
    read      = dataopts.get('read',None)
    nest      = dataopts.get('nest',True)
    ensure    = dataopts.get('ensure',True)
    overwrite = dataopts.get('overwrite',False)
    subinits  = dataopts.get('subinits',{})
    pathbase  = dataopts.get('pathbase')

    if overwrite and os.path.exists(workdir):
        shutil.rmtree(workdir)

    init_states = []


    initdir = os.path.join(pathbase or workdir,dataopts.get('initdir','init'))
    inputarchive = dataopts.get('inputarchive',None)
    if initdir:
        if inputarchive:
            prepare_workdir_from_archive(initdir, inputarchive, match = dataopts.get('archivematch'))
        init_state = LocalFSState(readonly = [initdir])
        init_states.append(init_state)

    def handle_init_spec(pathbase,spec):
        subinits = {}
        for k,v in spec.items():
            if isinstance(v,string_types):
                subinits[k] = {'_data': LocalFSState(readonly = [os.path.join(pathbase,v)])}
            if isinstance(v,dict):
                subinits.setdefault(k,{})['_data'] = LocalFSState(readonly = [os.path.join(pathbase,v.pop('_data'))]) if '_data' in v else None
                for kk,vv in handle_init_spec(pathbase,v).items():
                    subinits.setdefault(k,{})[kk] = vv
        return subinits


    if subinits:
        subinits = handle_init_spec(pathbase or workdir,subinits)

    writable_state = LocalFSState([workdir])
    return LocalFSProvider(read,writable_state,
        ensure = ensure, nest = nest, init_states = init_states,
        sub_init_states = subinits
    )



import contextlib
import importlib
import logging
import os
from adage.wflowcontroller import BaseController
from packtivity.syncbackends import defaultsyncbackend

from .reset import collective_downstream, remove_rules, reset_steps, undo_rules
from .wflow import YadageWorkflow
from .handlers.utils import handler_decorator

log = logging.getLogger(__name__)

ctrlhandlers, controller = handler_decorator()

class YadageController(BaseController):
    def __init__(self,*args, **kwargs):
        self.prepublishing_backend = defaultsyncbackend()
        self.disable_backend = False
        self.disable_prepublishing = kwargs.pop('disable_prepub',False)
        super(YadageController,self).__init__(*args,**kwargs)

    def sync_expected(self):
        for n in self.adageobj.dag.nodes():
            if 'YADAGE_IGNORE_PREPUBLISHING' in os.environ or self.disable_prepublishing:
                continue
            node = self.adageobj.dag.getNode(n)
            node.expected_result = self.prepublishing_backend.prepublish(
                node.task.spec, node.task.parameters.json(), node.task.state
            )

    def sync_backend(self):
        self.sync_expected()
        if not self.disable_backend:
            super(YadageController,self).sync_backend()

@controller('frommodel')
def frommodel_controller(ctrlstring, ctrlopts, model = None):
    if isinstance(model, YadageWorkflow):
        return YadageController(model,**ctrlopts)
    else:
        return PersistentController(model,**ctrlopts)

@controller('http')
def http_controller(ctrlstring, ctrlopts, model = None):
    try:
        from yadagehttpctrl.clientcontroller import YadageHTTPController
        ctrl = YadageHTTPController(server = ctrlstring, **ctrlopts)
        return ctrl
    except ImportError:
        log.exception('try installing yadagehttpctrl')

@controller('py:')
def frompython_controller(ctrlstring, ctrlopts, model = None):
    _, module, ctrlclass = ctrlstring.split(':')
    module = importlib.import_module(module)
    ctrlclass = getattr(module,ctrlclass)
    if ctrlopts.pop('pass_model',False):
        ctrlopts['model'] = model
    return ctrlclass(**ctrlopts)

def setup_controller(model = None, controller = 'frommodel', ctrlopts = None):
    '''
    return controller instance based on state configuration. For
    transaction-based states, returns PersistentController, for in-
    memory states returns in BaseController
    '''
    ctrlopts  = ctrlopts or {}
    for k in ctrlhandlers.keys():
        if controller.startswith(k):
            return ctrlhandlers[k](controller, ctrlopts, model)
    raise RuntimeError('unknown controller type %s' % controller)

class PersistentController(YadageController):
    '''
    workflow controller, that explicltly calls transaction methods on non read-only operations on the workflow state
    '''
    def __init__(self, model, backend = None):
        '''
        :param model: the model on whih the controller will operate
        :param backend: the backend to against which to check workflow state.

        :return: the controller instance
        '''
        self.model = model
        super(PersistentController, self).__init__(self.model.load(),backend)

    @contextlib.contextmanager
    def transaction(self, sync = True):
        '''the transaction context. will commit model to persistent store on exit.'''
        self.adageobj = self.model.load()
        if sync:
            log.debug('syncing to setup tx %s', self)
            super(PersistentController,self).sync_backend()
        yield

        isvalid = self.validate()
        if not isvalid:
            #raise RuntimeError('was about to commit invalid data!')
            log.warning('commit is invalid %s', isvalid)
        if sync:
            log.debug('syncing to teardown tx %s', self)
            super(PersistentController,self).sync_backend()
        self.model.commit(self.adageobj)

    def submit_nodes(self, nodeids):
        '''
        submit nodes to backend

        :param nodeids: list of ids of nodes to be submitted
        :return: None
        '''
        log.debug('transaction to submit')
        with self.transaction():
            nodes = [self.adageobj.dag.getNode(nodeid) for nodeid in nodeids]
            # log.info('submitting nodes to backend: %s', nodes)
            super(PersistentController,self).submit_nodes(nodes)

    def apply_rules(self, ruleids):
        '''
        apply rules to workflow instance

        :param ruleids: list of ids of reuls to be applied
        :return: None
        '''
        log.debug('transaction to apply')
        with self.transaction():
            rules = [r for r in self.adageobj.rules if r.identifier in ruleids]
            super(PersistentController,self).apply_rules(rules)

    def sync_backend(self):
        '''
        synchronize node data with backend
        '''
        log.debug('transaction to sync but (without sync in tx)')
        with self.transaction(sync = False): # disable sync to avoid infinite recursion
            super(PersistentController,self).sync_backend()

    def applicable_rules(self):
        '''
        :return: a list of rules whose predicate is fulfilled
        '''
        applicable_rules = super(PersistentController,self).applicable_rules()
        return [x.identifier for x in applicable_rules]

    def submittable_nodes(self):
        '''

        :return: a list of nodes with sucessfull and completed upstream
        '''
        submittable_nodes = super(PersistentController,self).submittable_nodes()
        return [x.identifier for x in submittable_nodes]

    def add_rules(self, rulespecs, dataarg, offset = '', groupname = None , dataopts = None):
        log.debug('adding %s rules', len(rulespecs))
        from .stages import JsonStage
        from .state_providers  import state_provider_from_string
        sp = state_provider_from_string(dataarg,dataopts)
        rules = [JsonStage(json, sp) for json in rulespecs]
        with self.transaction():
            self.adageobj.view(offset).addWorkflow(rules,groupname)

    def patch_rule(self, ruleid, patchspec):
        with self.transaction():
            rule = [x for x in self.adageobj.rules if x.identifier == ruleid][0]
            rule.rule.stagespec = patchspec

    def undo_rules(self, ruleids):
        '''
        :param nodeids: list of ids of nodes to reset
        :return: None
        '''
        with self.transaction():
            undo_rules(self.adageobj, ruleids)

    def remove_rules(self, ruleids):
        with self.transaction():
            remove_rules(self.adageobj, ruleids)

    def reset_nodes(self, nodeids):
        '''
        :param nodeids: list of ids of nodes to reset
        :return: None
        '''
        to_reset = nodeids + collective_downstream(self.adageobj, nodeids)
        with self.transaction():
            reset_steps(self.adageobj,to_reset)



import adage
from adage.serialize import obj_to_json

from .stages import JsonStage, OffsetStage
from .wflowview import WorkflowView
from .wflownode import YadageNode


class YadageWorkflow(adage.adageobject):
    '''
    The overall workflow state object that extends the basic
    Adage state object by two bookkeeping structures.
    '''

    def __init__(self, dag = None, rules = None, applied_rules = None, bookkeeping = None, stepsbystage = None, values = None):
        super(YadageWorkflow, self).__init__(
            dag = dag,
            rules = rules,
            applied_rules = applied_rules
        )
        self.stepsbystage = stepsbystage or {}
        self.bookkeeping = bookkeeping or {}
        self.values = values or {}

    def view(self, offset=''):
        return WorkflowView(self, offset)

    def json(self):
        json_or_nil = lambda x: None if x is None else x.json()
        data = obj_to_json(self,json_or_nil,json_or_nil)

        data['bookkeeping'] = self.bookkeeping
        data['stepsbystage'] = self.stepsbystage
        data['values'] = self.values
        return data

    @classmethod
    def fromJSON(cls, data,deserialization_opts = None, backend=None):
        def node_deserializer(data):
            node = YadageNode.fromJSON(data,deserialization_opts)
            if backend:
                # node.backend = backend
                node.update_state(backend = backend)
            return node

        def rule_deserializer(data):
            return OffsetStage.fromJSON(data,deserialization_opts)

        dag = adage.serialize.dag_from_json(
                    data['dag'],
                    node_deserializer
                )

        instance = cls(dag = dag,
            rules = [rule_deserializer(x) for x in data['rules'] ],
            applied_rules = [rule_deserializer(x) for x in data['applied'] ],
            bookkeeping = data['bookkeeping'],
            stepsbystage = data['stepsbystage'],
            values = data['values']
        )

        return instance

    @classmethod
    def createFromJSON(cls, jsondata, state_provider):
        instance = cls()
        rules = [JsonStage(stagedata, state_provider) for stagedata in jsondata['stages']]
        instance.view().addWorkflow(rules)
        return instance



import logging
import os
import copy

import adage

from .serialize import snapshot
from .wflowstate import load_model_fromstring
from .controllers import setup_controller
from .utils import setupbackend_fromstring, prepare_meta
from .creators import handlers as creators

log = logging.getLogger(__name__)

class YadageSteering(object):
    '''
    high level steering object to manage worklfow execution
    '''
    def __init__(self, metadir, controller):
        self.metadir = metadir
        self.controller = controller
        self.adage_kwargs = dict(workdir = os.path.join(metadir,'adage'))

    @classmethod
    def connect(cls, metadir, ctrlstring, ctrlopts = None, modelsetup = None, modelopts = None, accept_metadir = False):
        prepare_meta(metadir, accept_metadir)

        model = None
        if modelsetup:
            model = load_model_fromstring(modelsetup,modelopts)

        ctrl = setup_controller(
            model = model, controller = ctrlstring, ctrlopts = ctrlopts,
        )
        log.info('connected to model')
        return cls(metadir, ctrl)

    @classmethod
    def create(cls,**kwargs):
        dataarg = kwargs['dataarg']
        is_local_data = len(dataarg.split(':',1)) == 1
        if is_local_data:
            metadir = kwargs.get('metadir')
            metadir = metadir or '{}/_yadage/'.format(dataarg)
        else:
            metadir = kwargs['metadir']
        accept_metadir = kwargs.pop('accept_metadir', False)
        prepare_meta(metadir, accept_metadir)

        kw = copy.deepcopy(kwargs)
        kw['metadir'] = metadir

        ctrl = creators['local'](**kw)
        return cls(metadir, ctrl)

    @property
    def workflow(self):
        '''
        :return: the workflow object (from the controller)
        '''
        return self.controller.adageobj

    def adage_argument(self,**kwargs):
        '''
        add keyword arguments for workflow execution (adage)

        :param kwargs: adage keyword arguments (see adage documentation for options)
        '''
        self.adage_kwargs.update(**kwargs)

    def run_adage(self, backend = 'auto', **adage_kwargs):
        '''
        execution workflow with adage based against given backend
        :param backend: backend to use for packtivity processing.
        '''
        if backend=='auto':
            #respect if the controller already has a backend wired up
            self.controller.backend = self.controller.backend or setupbackend_fromstring('multiproc:auto')
            log.info('backend automatically set to %s', backend)
        elif backend:
            self.controller.backend = backend

        assert self.controller.backend
        self.adage_argument(**adage_kwargs)
        adage.rundag(controller = self.controller, **self.adage_kwargs)

    def serialize(self):
        '''
        serialized workflow and backend states (stored in meta directory)
        '''
        snapshot(
            self.workflow,
            '{}/yadage_snapshot_workflow.json'.format(self.metadir),
        )

    def visualize(self):
        '''
        generate workflow visualization (stored in meta directory)
        '''
        import yadage.visualize as visualize
        visualize.write_prov_graph(self.metadir, self.workflow, vizformat='png')
        visualize.write_prov_graph(self.metadir, self.workflow, vizformat='pdf')



import os
import importlib
import logging
from .localposix import LocalFSProvider

log = logging.getLogger(__name__)

from ..handlers.utils import handler_decorator

provider_deserialize_handlers, provider_deserializer = handler_decorator()

@provider_deserializer('localfs_provider')
def localfs_provider_deserializer(jsondata, deserialization_opts):
    return LocalFSProvider.fromJSON(jsondata, deserialization_opts)

@provider_deserializer('frompython_provider')
def frompython_provider(jsondata, deserialization_opts):
    providerstring = deserialization_opts.get('state_provider','')
    _, module, providerclass = providerstring.split(':')
    module = importlib.import_module(module)
    providerclass = getattr(module,providerclass)
    provideropts = {}
    return providerclass.fromJSON(jsondata,**provideropts)

@provider_deserializer('fromenv_provider')
def fromenv_provider_deserializer(jsondata, deserialization_opts):
    module = importlib.import_module(os.environ['YADAGE_STATEPROVIDER'])
    return module.load_provider(jsondata)

def load_provider(jsondata,deserialization_opts = None):
    log.debug('load_provider opts %s', deserialization_opts)
    deserialization_opts = deserialization_opts or {}
    if jsondata == None:
        return None
    if 'state_provider' in deserialization_opts:
        providerstring = deserialization_opts.get('state_provider','')
        if providerstring.startswith('py:'):
            return provider_deserialize_handlers['frompython_provider'](jsondata, deserialization_opts)
    if 'YADAGE_STATEPROVIDER' in os.environ:
        return provider_deserialize_handlers['fromenv_provider'](jsondata, deserialization_opts)

    return provider_deserialize_handlers[jsondata['state_provider_type']](jsondata, deserialization_opts)

    raise TypeError('unknown provider type {}'.format(jsondata['state_provider_type']))

providersetup_handlers, providersetup = handler_decorator()

@providersetup('local')
def localfs_provider(dataarg,dataopts):
    import yadage.state_providers.localposix
    return yadage.state_providers.localposix.setup_provider(dataarg,dataopts)

@providersetup('py:')
def formpython_provider(dataarg,dataopts):
    _,module, setupfunc,dataarg = dataarg.split(':',3)
    module = importlib.import_module(module)
    setupfunc = getattr(module,setupfunc)
    return setupfunc(dataarg,dataopts)

@providersetup('fromenv:')
def fromenv_provider(dataarg,dataopts):
    module = importlib.import_module(os.environ['PACKTIVITY_STATEPROVIDER'])
    return module.setup_provider(dataarg,dataopts)

def state_provider_from_string(dataarg,dataopts = None):
    dataopts = dataopts or {}
    if len(dataarg.split(':',1)) == 1:
        dataarg = 'local:'+dataarg
    for k in providersetup_handlers.keys():
        if dataarg.startswith(k):
            return providersetup_handlers[k](dataarg,dataopts)
    raise RuntimeError('unknown data type %s %s' % (dataarg, dataopts))
